/* The GIMP -- an image manipulation program
 * Copyright (C) 1995-1997 Spencer Kimball and Peter Mattis
 *
 * GIMP Plug-in for Texas Instruments Pictures files.
 * Copyright (C) 2002 Romain Lievin <roms@lpg.ticalc.org>
 * Some parts from Christian Kreibich's WinIcon plugin <christian@whoop.org>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <libgimp/gimpdrawable.h>

#define PIC_DBG

#include "callbacks.h"
#include "picdialog.h"
#include "plugin-intl.h"
#include "main.h"

static GtkWidget *pic_preview_new(gint32 layer);
static void       pic_fill_preview_with_thumb (GtkWidget *widget, gint32 drawable_ID);


static GtkWidget *
pic_preview_new(gint32 layer)
{
  GtkWidget *pict_preview;

  pict_preview = gtk_preview_new (GTK_PREVIEW_COLOR);
  pic_fill_preview_with_thumb (pict_preview, layer);

  return pict_preview;
}


static void
pic_fill_preview_with_thumb (GtkWidget *widget, 
			     gint32     drawable_ID)
{
  guchar  *drawable_data;
  gint     bpp;
  gint     x,y;
  gint     width;
  gint     height;
  guchar  *src;
  gdouble  r, g, b, a;
  gdouble  c0, c1;
  guchar  *p0, *p1;
  guchar  *even, *odd;

  width = gimp_drawable_width(drawable_ID);
  height = gimp_drawable_height(drawable_ID);
  bpp = 0;
  
  if (width > 128)
    width = 128;
  if (height > 128)
    height = 128;

  drawable_data = 
    gimp_drawable_get_thumbnail_data (drawable_ID, &width, &height, &bpp);

  if (width < 1 || height < 1)
    return;

  gtk_preview_size (GTK_PREVIEW (widget), width, height);

  even = g_malloc (width * 3);
  odd  = g_malloc (width * 3);
  src = drawable_data;

  for (y = 0; y < height; y++)
    {
      p0 = even;
      p1 = odd;
      
      for (x = 0; x < width; x++) 
	{
	  if(bpp == 4)
	    {
	      r = ((gdouble)src[x*4+0])/255.0;
	      g = ((gdouble)src[x*4+1])/255.0;
	      b = ((gdouble)src[x*4+2])/255.0;
	      a = ((gdouble)src[x*4+3])/255.0;
	    }
	  else if(bpp == 3)
	    {
	      r = ((gdouble)src[x*3+0])/255.0;
	      g = ((gdouble)src[x*3+1])/255.0;
	      b = ((gdouble)src[x*3+2])/255.0;
	      a = 1.0;
	    }
	  else
	    {
	      r = ((gdouble)src[x*bpp+0])/255.0;
	      g = b = r;
	      if(bpp == 2)
		a = ((gdouble)src[x*bpp+1])/255.0;
	      else
		a = 1.0;
	    }
	  
	  if ((x / GIMP_CHECK_SIZE_SM) & 1) 
	    {
	      c0 = GIMP_CHECK_LIGHT;
	      c1 = GIMP_CHECK_DARK;
	    } 
	  else 
	    {
	      c0 = GIMP_CHECK_DARK;
	      c1 = GIMP_CHECK_LIGHT;
	    }
	  
	*p0++ = (c0 + (r - c0) * a) * 255.0;
	*p0++ = (c0 + (g - c0) * a) * 255.0;
	*p0++ = (c0 + (b - c0) * a) * 255.0;
	
	*p1++ = (c1 + (r - c1) * a) * 255.0;
	*p1++ = (c1 + (g - c1) * a) * 255.0;
	*p1++ = (c1 + (b - c1) * a) * 255.0;
	
      }
      
      if ((y / GIMP_CHECK_SIZE_SM) & 1)
	gtk_preview_draw_row (GTK_PREVIEW (widget), (guchar *)odd,  0, y, width);
      else
	gtk_preview_draw_row (GTK_PREVIEW (widget), (guchar *)even, 0, y, width);

      src += width * bpp;
    }

  g_free (even);
  g_free (odd);
  g_free (drawable_data);
}


/* This function creates and returns an hbox for an pict,
   which then gets added to the dialog's main vbox. */
static GtkWidget*
pic_create_pict_hbox (GtkWidget *pict_preview,
		      gint32 layer, gint layer_num)
{
  GtkWidget *pict_hbox;
  GtkWidget *pict_optionmenu;
  GtkWidget *pict_optionmenu_menu;
  GtkWidget *glade_menuitem;

  pict_hbox = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (pict_hbox);
  gtk_widget_show (pict_hbox);

  /* To make life easier for the callbacks, we store the
     layer's ID and stacking number with the hbox. */

  gtk_object_set_data (GTK_OBJECT (pict_hbox), "pict_layer", GINT_TO_POINTER(layer));
  gtk_object_set_data (GTK_OBJECT (pict_hbox), "pict_layer_num", GINT_TO_POINTER(layer_num));

  gtk_object_set_data_full (GTK_OBJECT (pict_hbox), "pict_preview", pict_preview,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (pict_preview);
  gtk_box_pack_start (GTK_BOX (pict_hbox), pict_preview, TRUE, TRUE, 0);

  pict_optionmenu = gtk_option_menu_new ();
  gtk_widget_ref (pict_optionmenu);
  gtk_object_set_data_full (GTK_OBJECT (pict_hbox), "pict_optionmenu", pict_optionmenu,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (pict_optionmenu);
  gtk_box_pack_start (GTK_BOX (pict_hbox), pict_optionmenu, FALSE, FALSE, 0);
  pict_optionmenu_menu = gtk_menu_new ();


  glade_menuitem = gtk_menu_item_new_with_label (_("TI73 (96x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti73_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI82 (96x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti82_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI83 (96x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti83_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI83+ (96x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti83p_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI85 (128x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti85_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI86 (128x63 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti86_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI89 (239x103 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti89_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI92 (239x103 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti92_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("TI92+ (239x103 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_ti92p_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);

  glade_menuitem = gtk_menu_item_new_with_label (_("V200 (239x103 pixels)"));
  gtk_signal_connect (GTK_OBJECT (glade_menuitem), "activate",
		      on_v200_menu_selected, pict_hbox);
  gtk_widget_show (glade_menuitem);
  gtk_menu_append (GTK_MENU (pict_optionmenu_menu), glade_menuitem);
  

  gtk_option_menu_set_menu (GTK_OPTION_MENU (pict_optionmenu), pict_optionmenu_menu);
  gtk_option_menu_set_history (GTK_OPTION_MENU (pict_optionmenu), 3);

  return pict_hbox;
}


GtkWidget*       
pic_specs_dialog_new (gint num_layers)
{
  GtkWidget *dialog1;
  GtkWidget *dialog_vbox1;
  GtkWidget *picts_frame;
  GtkWidget *picts_scrolledwindow;
  GtkWidget *viewport1;
  GtkWidget *picts_vbox;
  GtkWidget *dialog_action_area1;
  GtkWidget *hbox2;
  gint      *calc_type, i;

  dialog1 = gimp_dialog_new (_("GIMP Ti Picture Plugin"), PLUGIN_NAME,
			     gimp_standard_help_func, "filters/"PLUGIN_NAME".html",
			     GTK_WIN_POS_MOUSE,
			     FALSE, FALSE, FALSE,
			     
			     _("OK"), on_ok_button_clicked,
			     NULL, NULL, NULL, TRUE, FALSE,

			     _("Cancel"), on_cancel_button_clicked,
			     NULL, NULL, NULL, FALSE, TRUE,			     

			     _("Help"), gimp_standard_help_func,
			     NULL, "filters/"PLUGIN_NAME".html", NULL, FALSE, TRUE,

			     NULL);


  gtk_object_set_data (GTK_OBJECT (dialog1), "dialog1", dialog1);
  gtk_window_set_title (GTK_WINDOW (dialog1), _("Picture Size Dialog"));
  gtk_window_set_policy (GTK_WINDOW (dialog1), FALSE, TRUE, FALSE);

  /* We store an array that holds each pict's requested size
     with the dialog. It's queried when the dialog is closed so the
     save routine knows what colormaps etc to generate in the saved
     file. We store twice the number necessary because in the second
     set, the color depths that are automatically suggested are stored
     for later comparison.
  */

  calc_type = g_new(gint, 2 * num_layers);
  for (i = 0; i < 2 * num_layers; i++)
    calc_type[i] = CALC_NONE;

  gtk_object_set_data (GTK_OBJECT(dialog1), "calc_type", calc_type);

  dialog_vbox1 = GTK_DIALOG (dialog1)->vbox;
  gtk_object_set_data (GTK_OBJECT (dialog1), "dialog_vbox1", dialog_vbox1);
  gtk_widget_show (dialog_vbox1);

  picts_frame = gtk_frame_new (_("Picture Details"));
  gtk_widget_ref (picts_frame);
  gtk_object_set_data_full (GTK_OBJECT (dialog1), "picts_frame", picts_frame,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (picts_frame);
  gtk_box_pack_start (GTK_BOX (dialog_vbox1), picts_frame, TRUE, TRUE, 0);
  gtk_container_set_border_width (GTK_CONTAINER (picts_frame), 5);

  picts_scrolledwindow = gtk_scrolled_window_new (NULL, NULL);
  gtk_widget_ref (picts_scrolledwindow);
  gtk_object_set_data_full (GTK_OBJECT (dialog1), "picts_scrolledwindow", picts_scrolledwindow,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (picts_scrolledwindow);
  gtk_container_add (GTK_CONTAINER (picts_frame), picts_scrolledwindow);
  gtk_scrolled_window_set_policy (GTK_SCROLLED_WINDOW (picts_scrolledwindow), GTK_POLICY_AUTOMATIC, GTK_POLICY_AUTOMATIC);
  gtk_container_set_border_width (GTK_CONTAINER (picts_scrolledwindow), 5);

  viewport1 = gtk_viewport_new (NULL, NULL);
  gtk_widget_ref (viewport1);
  gtk_object_set_data_full (GTK_OBJECT (dialog1), "viewport1", viewport1,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (viewport1);
  gtk_container_add (GTK_CONTAINER (picts_scrolledwindow), viewport1);
  gtk_container_set_border_width (GTK_CONTAINER (viewport1), 0);

  picts_vbox = gtk_vbox_new (TRUE, 0);
  gtk_widget_ref (picts_vbox);
  gtk_object_set_data_full (GTK_OBJECT (dialog1), "picts_vbox", picts_vbox,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (picts_vbox);
  gtk_container_add (GTK_CONTAINER (viewport1), picts_vbox);

  /* content used to be added here ... */

  dialog_action_area1 = GTK_DIALOG (dialog1)->action_area;
  gtk_object_set_data (GTK_OBJECT (dialog1), "dialog_action_area1", dialog_action_area1);
  gtk_widget_show (dialog_action_area1);
  gtk_container_set_border_width (GTK_CONTAINER (dialog_action_area1), 5);

  hbox2 = gtk_hbox_new (FALSE, 0);
  gtk_widget_ref (hbox2);
  gtk_object_set_data_full (GTK_OBJECT (dialog1), "hbox2", hbox2,
                            (GtkDestroyNotify) gtk_widget_unref);
  gtk_widget_show (hbox2);
  gtk_box_pack_start (GTK_BOX (dialog_action_area1), hbox2, TRUE, TRUE, 0);

  return dialog1;
}

GtkWidget *
pic_specs_dialog_get_layer_preview(GtkWidget *dialog, gint32 layer)
{
  GtkWidget *preview;
  GtkWidget *pict_hbox;
  gchar      key[MAXLEN];

  g_snprintf(key, MAXLEN, "layer_%i_hbox", layer);
  pict_hbox = gtk_object_get_data(GTK_OBJECT(dialog), key);

  if (!pict_hbox)
    {
      D(("Something's wrong -- couldn't look up hbox by layer ID\n"));
      return NULL;
    }

  preview = gtk_object_get_data(GTK_OBJECT(pict_hbox), "pict_preview");

  if (!pict_hbox)
    {
      D(("Something's wrong -- couldn't look up preview from hbox\n"));
      return NULL;
    }

  return preview;
}


void             
pic_specs_dialog_add_pict (GtkWidget *dialog, gint32 layer, gint layer_num)
{
  GtkWidget *picts_vbox;
  GtkWidget *pict_hbox;
  GtkWidget *pict_preview;
  gchar      key[MAXLEN];

  picts_vbox = gtk_object_get_data (GTK_OBJECT(dialog), "picts_vbox");
  pict_preview = pic_preview_new (layer);  
  pict_hbox = pic_create_pict_hbox (pict_preview, layer, layer_num);
  gtk_box_pack_start (GTK_BOX(picts_vbox), pict_hbox, TRUE, FALSE, 2);

  /* Let's make the hbox accessible through the layer ID */
  g_snprintf(key, MAXLEN, "layer_%i_hbox", layer);
  gtk_object_set_data (GTK_OBJECT(dialog), key, pict_hbox);
}


void             
pic_specs_dialog_update_pict_preview(GtkWidget *dialog, gint32 layer,
				     gint calc_type)
{
  GtkWidget      *preview;
  GimpPixelRgn    src_pixel_rgn, dst_pixel_rgn;
  gint32          tmp_image;
  gint32          tmp_layer;
  gint            w, h;
  guchar         *buffer;
  gboolean        result;

  tmp_image = gimp_image_new(gimp_drawable_width(layer),
			     gimp_drawable_height(layer),
			     GIMP_RGB);
  
  w = gimp_drawable_width(layer);
  h = gimp_drawable_height(layer);

  tmp_layer = gimp_layer_new(tmp_image, "tmp", w, h,
                              GIMP_RGBA_IMAGE, 100, GIMP_NORMAL_MODE);
  gimp_pixel_rgn_init(&src_pixel_rgn, gimp_drawable_get(layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_init(&dst_pixel_rgn, gimp_drawable_get(tmp_layer),
		       0, 0, w, h, TRUE, FALSE);
  buffer = g_malloc(w * h * sizeof(gint32));
  gimp_pixel_rgn_get_rect(&src_pixel_rgn, buffer, 0, 0, w, h);
  gimp_pixel_rgn_set_rect(&dst_pixel_rgn, buffer, 0, 0, w, h);
  gimp_image_add_layer(tmp_image, tmp_layer, 0);

  result = gimp_image_convert_indexed(tmp_image,
				      GIMP_FS_DITHER,
				      GIMP_MAKE_PALETTE,
				      2,
				      TRUE,
				      FALSE,
				      "dummy");

  preview = pic_specs_dialog_get_layer_preview(dialog, layer);
  pic_fill_preview_with_thumb(preview, tmp_layer);
  gtk_widget_queue_draw(preview);

  gimp_image_delete(tmp_image);
  g_free(buffer);
}

