/* The GIMP -- an image manipulation program
 * Copyright (C) 1995-1997 Spencer Kimball and Peter Mattis
 *
 * GIMP Plug-in for Texas Instruments Pictures files.
 * Copyright (C) 2002 Romain Lievin <roms@lpg.ticalc.org>
 * Some parts from Christian Kreibich's WinIcon plugin <christian@whoop.org>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"

#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <tilp/tifiles.h>

#define PIC_DBG

#include "main.h"
#include "plugin-intl.h"


/*
  TI89,92,92+,V200: 239x103 pixels images (stored as 103 rows of 240 pixels).
   0   2  : number of bytes minus 2 ?
   2   2  : height in pixels
   4   2  : width in pixels
   6   n  : bitmap
   6+n 1  : always DFh
  TI73,82,83,83+: 96x63 pixels images.
   0   2  : number of bytes
   2   n  : bitmap
  TI85,86: 128x63 pixels images.
   0   2  : number of bytes
   2   n  : bitmap
*/


static gboolean
pic_init (char *filename, TiPicture *pic)
{
  memset(pic, 0, sizeof(TiPicture));

  if (! (pic->fp = fopen (filename, "r")))
    return FALSE;

  pic->filename = filename;

  pic->content = tifiles_create_regular_content();

  return TRUE;
}

static gboolean
pic_load (TiPicture *pic)
{
  gint err;
  const gchar *desc;
  gint calc_type;

  if (!pic)
    return TRUE;

  if(!tifiles_is_a_single_file(pic->filename))
    return FALSE;

  desc = tifiles_file_descriptive(pic->filename);
  if(strcmp(desc, "Picture"))
    return FALSE;

  calc_type = tifiles_which_calc_type(pic->filename);

  err = tifiles_read_regular_file(pic->filename, pic->content); 
  if(err)
    {
      char str[256];
      tifiles_get_error(err, str);
      D(("%s", str));
      return FALSE;
    }

  pic->entry = &(pic->content->entries[0]);

  switch(calc_type)
    {
    case CALC_TI73: case CALC_TI82: case CALC_TI83: case CALC_TI83P:
      pic->nbytes = (pic->entry->data[1] << 8) | pic->entry->data[0];      
      pic->width  = 96;
      pic->height = 63;
      pic->bitmap = pic->entry->data + 2;
      break;
    case CALC_TI85: case CALC_TI86:
      pic->nbytes =(pic->entry->data[1] << 8) | pic->entry->data[0];
      pic->width  = 128;
      pic->height = 63;
      pic->bitmap = pic->entry->data + 2;
      break;
    case CALC_TI89: case CALC_TI92: case CALC_TI92P: case CALC_V200:
      pic->nbytes =(pic->entry->data[0] << 8) | pic->entry->data[1];
      pic->nbytes -= 5;
      pic->height =(pic->entry->data[2] << 8) | pic->entry->data[3];
      pic->width  =(pic->entry->data[4] << 8) | pic->entry->data[5]; 
      pic->bitmap = pic->entry->data + 6;
      break;
    default:
      return FALSE;
    }
  
  D(("*** %s: TI picture file, %ix%i (%i bytes)\n", pic->filename,
	 pic->width, pic->height, pic->nbytes));

  return FALSE;
}

static gint32
pic_to_gimp(TiPicture *pic)
{
  gint            x, y;
  gint            w, h, s;
  gint32          image;
  gint32          layer;
  guint8         *rgb;
  GimpDrawable   *drawable;
  GimpPixelRgn    pixel_rgn;

  w = pic->width;
  h = pic->height;
  s = pic->nbytes / pic->height;
  D(("w,h,s = %i,%i,%i\n", w, h, s));

  // Allocate the Gimp image
  image = gimp_image_new(w, h, GIMP_RGB);
  gimp_image_set_filename(image, pic->filename);

  layer = gimp_layer_new(image, "Ti Picture", w, h,
			  GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
  gimp_image_add_layer(image, layer, 0);
  drawable = gimp_drawable_get(layer);

  // copy data (bitmap to rgb conversion)
  rgb = g_malloc0(3 * w * h);

  for(y = 0; y < h; y++)
    for(x = 0; x < w; x++)
      {
	int bit;

	bit = pic->bitmap[s*y + x/8] &(128 >>(x&7)) ? 1 : 0;

	if(!bit)
	  {
	    rgb[3*(w*y + x) + 0] = 0xff;
	    rgb[3*(w*y + x) + 1] = 0xff;
	    rgb[3*(w*y + x) + 2] = 0xff;
	  }
	else
	  {
	    rgb[3*(w*y + x) + 0] = 0x00;
	    rgb[3*(w*y + x) + 1] = 0x00;
	    rgb[3*(w*y + x) + 2] = 0x00;
	  }
      }
  
  gimp_pixel_rgn_init(&pixel_rgn, drawable,
		      0, 0, drawable->width, drawable->height, TRUE, FALSE);
  gimp_pixel_rgn_set_rect(&pixel_rgn, (guint8 *)rgb,
			  0, 0, drawable->width, drawable->height);
  
  g_free(rgb);
  gimp_drawable_flush(drawable);
  gimp_drawable_detach(drawable);
  
  gimp_image_set_active_layer(image, 0);

  if(interactive_pic)
    gimp_progress_update(1);

  return image;
}

gint32
LoadPIC(gchar *filename)
{
  gint32 image_ID;
  TiPicture pic;

  if(interactive_pic)
    {
      guint8 *temp_buf = g_strdup_printf(_("Loading %s:"), filename);
      gimp_progress_init(temp_buf);
      g_free(temp_buf);
    }

  if(!pic_init(filename, &pic))
    return -1;

  pic_load(&pic);

  if((image_ID = pic_to_gimp(&pic)) < 0)
    {
      pic_cleanup(&pic);
      return -1;
    }

  pic_cleanup(&pic);

  D(("*** picture successfully loaded.\n\n"));
  
  return image_ID;
}
