/* The GIMP -- an image manipulation program
 * Copyright (C) 1995-1997 Spencer Kimball and Peter Mattis
 *
 * GIMP Plug-in for Texas Instruments Pictures files.
 * Copyright (C) 2002 Romain Lievin <roms@lpg.ticalc.org>
 * Some parts from Christian Kreibich's WinIcon plugin <christian@whoop.org>.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "config.h"
#include <gtk/gtk.h>
#include <libgimp/gimp.h>
#include <libgimp/gimpui.h>
#include <string.h>
#include <stdlib.h>

#define PIC_DBG

#include "main.h"
#include "plugin-intl.h"
#include "picdialog.h"

/* */

static gboolean
pic_layers_too_big(gint32 image)
{
  gint *layers;
  gint  i, num_layers;

  layers = gimp_image_get_layers(image, &num_layers);

  for(i = 0; i < num_layers; i++)
    {
      if((gimp_drawable_width(layers[i]) > 255) ||
	 (gimp_drawable_height(layers[i]) > 255))
	{
	  g_free(layers);
	  return TRUE;
	}
    }

  return FALSE;
}

static void
pic_free_color_item(gpointer data1, gpointer data2, gpointer data3)
{
  g_free(data1);
  data2 = NULL; // Shut up warnings
  data3 = NULL; // Shut up warnings
}

static gint
pic_get_layer_num_colors(gint32 layer)
{
  GimpPixelRgn    pixel_rgn;
  gint            x, y, w, h;
  gint            num_colors = 0;
  guint8         *buffer = NULL;
  GHashTable     *hash;
  
  w = gimp_drawable_width(layer);
  h = gimp_drawable_height(layer);
  buffer = g_malloc(3 * w * h);
  
  gimp_pixel_rgn_init(&pixel_rgn, gimp_drawable_get(layer),
		      0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_get_rect(&pixel_rgn, (guint8 *) buffer, 0, 0, w, h);
  
  hash = g_hash_table_new(g_int_hash, g_int_equal);
  
  for(y = 0; y < h; y++)
    for(x = 0; x < w; x++)
      {
	guint8  r, g, b;
	guint32 *color;
	
	color = g_new0(guint32, 1);
	r = buffer[3*(y*w + x) + 0];
	g = buffer[3*(y*w + x) + 1];
	b = buffer[3*(y*w + x) + 2];
	
	*color = (r << 16) | (g << 8) | b;

	g_hash_table_insert(hash, color, color);
      }

  num_colors = g_hash_table_size(hash);

  g_hash_table_foreach(hash, pic_free_color_item, NULL);
  g_hash_table_destroy(hash);
  g_free(buffer);

  return num_colors;
}

static void      
pic_image_get_reduced_buf(guint32 layer, guchar **buf_out, gint calc_type)
{
  GimpPixelRgn    src_pixel_rgn, dst_pixel_rgn;
  gint32          tmp_image;
  gint32          tmp_layer;
  gint            w, h;
  guint8         *buffer;
  guchar         *cmap;
  gint            num_colors;

  w = gimp_drawable_width(layer);
  h = gimp_drawable_height(layer);

  buffer = g_malloc(3 * w * h);

  tmp_image = gimp_image_new(w, h, GIMP_RGB);
  tmp_layer = gimp_layer_new(tmp_image, "tmp", w, h,
			      GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
  gimp_pixel_rgn_init(&src_pixel_rgn, gimp_drawable_get(layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_init(&dst_pixel_rgn, gimp_drawable_get(tmp_layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_get_rect(&src_pixel_rgn, buffer, 0, 0, w, h);
  gimp_pixel_rgn_set_rect(&dst_pixel_rgn, buffer, 0, 0, w, h);
  gimp_image_add_layer(tmp_image, tmp_layer, 0);
      
  gimp_image_convert_indexed(tmp_image,
			     GIMP_FS_DITHER,
			     GIMP_MAKE_PALETTE,
			     2,
			     TRUE,
			     FALSE,
			     "dummy");

  cmap = gimp_image_get_cmap(tmp_image, &num_colors);
  D(("#colors = %i\n", num_colors));
  D(("cmap: %3i %3i %3i\n", cmap[3*0+0], cmap[3*0+1], cmap[3*0+2]));
  D(("cmap: %3i %3i %3i\n", cmap[3*1+0], cmap[3*1+1], cmap[3*1+2]));

  gimp_image_convert_rgb(tmp_image);
  
  gimp_pixel_rgn_init(&dst_pixel_rgn, gimp_drawable_get(tmp_layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_get_rect(&dst_pixel_rgn, buffer, 0, 0, w, h);
  
  gimp_image_delete(tmp_image);
  
  *buf_out = buffer;
}

static void      
pic_image_reduce_layer_size(guint32 layer, gint calc_type)
{
  GimpPixelRgn    src_pixel_rgn, dst_pixel_rgn;
  gint32          tmp_image;
  gint32          tmp_layer;
  gint            w, h;
  guchar         *buffer;

  w = gimp_drawable_width(layer);
  h = gimp_drawable_height(layer);

  buffer = g_malloc(3 * w * h);

  tmp_image = gimp_image_new(w, h, GIMP_RGB);
  tmp_layer = gimp_layer_new(tmp_image, "tmp", w, h,
			      GIMP_RGB_IMAGE, 100, GIMP_NORMAL_MODE);
  gimp_pixel_rgn_init(&src_pixel_rgn, gimp_drawable_get(layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_init(&dst_pixel_rgn, gimp_drawable_get(tmp_layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_get_rect(&src_pixel_rgn, buffer, 0, 0, w, h);
  gimp_pixel_rgn_set_rect(&dst_pixel_rgn, buffer, 0, 0, w, h);
  gimp_image_add_layer(tmp_image, tmp_layer, 0);
  
  gimp_image_convert_indexed(tmp_image,
			     GIMP_FS_DITHER,
			     GIMP_MAKE_PALETTE,
			     2,
			     TRUE,
			     FALSE,
			     "dummy");
      
  gimp_image_convert_rgb(tmp_image);
  
  gimp_pixel_rgn_init(&dst_pixel_rgn, gimp_drawable_get(tmp_layer),
		       0, 0, w, h, TRUE, FALSE);
  gimp_pixel_rgn_get_rect(&dst_pixel_rgn, buffer, 0, 0, w, h);
  gimp_pixel_rgn_set_rect(&src_pixel_rgn, buffer, 0, 0, w, h);
  
  gimp_image_delete(tmp_image);
  
  gimp_drawable_update(layer, 0, 0, w ,h);
  gimp_drawable_flush(gimp_drawable_get(layer));
  gimp_displays_flush();
  
  g_free(buffer);
}

static void
pic_sync_image_to_pic(gint32 image, gint *calc_type)
{
  gint *layers;
  gint  i, num_layers;

  layers = gimp_image_get_layers(image, &num_layers);

  for(i = 0; i < num_layers; i++)
    {
      D(("Layer '%s' was reduced -- updating source image.\n",
	 gimp_layer_get_name(layers[i])));
      pic_image_reduce_layer_size(layers[i], calc_type[i]);
    }

  g_free(layers);
}

/* */



static GimpPDBStatusType
pic_init(char *filename, TiPicture *pic, gint *calc_typ)
{
  gint calc_type = calc_typ[0];

  memset(pic, 0, sizeof(TiPicture));

  if(!(pic->fp = fopen(filename, "wb")))
    {
      g_message(_("Can't open %s"), filename);
      return GIMP_PDB_EXECUTION_ERROR;
    }

  tifiles_set_calc(calc_type);
  switch(calc_type)
    {
    case CALC_TI73: case CALC_TI82: case CALC_TI83: case CALC_TI83P:
      pic->width  = 96;
      pic->height = 63;
      pic->offset = 2;
      break;
    case CALC_TI85: case CALC_TI86:
      pic->width  = 128;
      pic->height = 63;
      pic->offset = 2;
      break;
    case CALC_TI89: case CALC_TI92: case CALC_TI92P: case CALC_V200:
      pic->width  = 239;
      pic->height = 103;
      pic->offset = 6;
      break;
    default:
      return FALSE;
    }
  pic->nbytes = (ceilf(1.0 * pic->width / 8) * pic->height);
  D(("%i-%i-%i\n", pic->width, pic->height, pic->nbytes));

  // Initialize TI picture structure
  pic->filename = filename;
  pic->content  = tifiles_create_regular_content();

  // Initialize TiRegular structure
  pic->content->calc_type = calc_type;
  strcpy(pic->content->comment, "Picture file written by 'The GIMP'");
  pic->content->num_entries = 1;
  pic->content->entries =(TiVarEntry *)tifiles_calloc(1, sizeof(TiVarEntry));
  
  // Set up entry alias
  pic->entry = &(pic->content->entries[0]);

  // Initialize TiVarEntry structure
  strcpy(pic->entry->folder, "main");
  strncpy(pic->entry->name, filename, 8); pic->entry->name[9] = '\0';
  pic->entry->type = tifiles_string2vartype("PIC");
  pic->entry->attr = ATTRB_NONE;
  pic->entry->size = pic->nbytes + pic->offset;
  pic->entry->data =(uint8_t *)tifiles_calloc(pic->entry->size, 1);

  // Set up bitmap alias
  pic->bitmap = pic->entry->data + pic->offset;

  return GIMP_PDB_SUCCESS;
}

static void
pic_setup(TiPicture *pic, gint32 image_ID, gint *calc_type)
{ 
  gint   *layers, num_layers, layer;
  gint   x, y;
  gint   w, h, s;
  gint   bit;
  gint   n;
  guint8 *buffer = NULL;

  // get image size
  layers = gimp_image_get_layers(image_ID, &num_layers);
  layer = layers[0];
  w = pic->width;  //gimp_drawable_width(layer);
  h = pic->height; //gimp_drawable_height(layer);
  s = pic->nbytes / pic->height;
  D(("w,h,s = %i,%i,%i\n", w, h, s));

  // reduce # of colors
  n = pic_get_layer_num_colors(layer);
  pic_image_get_reduced_buf(layer, &buffer, *calc_type);

  // copy data (rgb to bitmap conversion)
  for(y = 0; y < h; y++)
    for(x = 0; x < w; x += 8)
      {
	guint8 mask = 0x00;
	
	for(bit=0; bit<8; bit++)
	  {
	    guint8 data = buffer[3*(w*y + x+bit) + 0];

	    mask <<= 1;
	    if(!data) 
	      mask |= 1;
	  }
	pic->bitmap[s*y + x/8] = mask;
      }

  switch(calc_type[0])
    {
    case CALC_TI73: case CALC_TI82: case CALC_TI83: case CALC_TI83P:
      pic->entry->data[0] = LSB(pic->nbytes + pic->offset - 2);
      pic->entry->data[1] = MSB(pic->nbytes + pic->offset - 2);
      break;
    case CALC_TI85: case CALC_TI86:
      pic->entry->data[0] = LSB(pic->nbytes + pic->offset - 2);
      pic->entry->data[1] = MSB(pic->nbytes + pic->offset - 2);
      break;
    case CALC_TI89: case CALC_TI92: case CALC_TI92P: case CALC_V200:
      pic->entry->data[0] = MSB(pic->nbytes + pic->offset - 2 + 1);
      pic->entry->data[1] = LSB(pic->nbytes + pic->offset - 2 + 1);
      pic->entry->data[2] = MSB(pic->height);
      pic->entry->data[3] = LSB(pic->height);
      pic->entry->data[4] = MSB(pic->width);
      pic->entry->data[5] = LSB(pic->width);
      break;
    default:
      return;
    } 
}

static GimpPDBStatusType
pic_save(TiPicture *pic)
{ 
  int err;

  err = tifiles_write_regular_file(pic->filename, pic->content, NULL);
  if(err)
    return GIMP_PDB_EXECUTION_ERROR;

  return GIMP_PDB_SUCCESS;
}

static gint*
pic_show_pict_dialog(gint32 image_ID)
{
  GtkWidget *dialog, *hbox;
  GtkWidget *pict_optionmenu;
  gint      *layers;
  gint       i, num_layers;
  gint       exit_okay;
  gchar      key[MAXLEN];
  gint      *calc_type = NULL;

  // Init UI
  gimp_ui_init(PLUGIN_NAME, TRUE);

  // Initialize tooltips
  gimp_help_init();

  layers = gimp_image_get_layers(image_ID, &num_layers);
  dialog = pic_specs_dialog_new(num_layers);

  for(i = 0; i < num_layers; i++)
    {
      /* if(gimp_layer_get_visible(layers[i])) */
      pic_specs_dialog_add_pict(dialog, layers[i], i);
    }

  // Scale the thing to approximately fit its content, but not too large ...
  gtk_widget_realize(dialog);
  gtk_window_set_default_size(GTK_WINDOW(dialog), 400, 200);

  calc_type = gtk_object_get_data(GTK_OBJECT(dialog), "calc_type");

  /* Limit the color depths to values that don't cause any color loss --
     the user should pick these anyway, so we can save her some time.
     If the user wants to lose some colors, the settings can always be changed
     in the dialog: */
  for(i = 0; i < num_layers; i++)
    {
      gint w, h;
      w = gimp_drawable_width(layers[i]);
      h = gimp_drawable_height(layers[i]);

      g_snprintf(key, MAXLEN, "layer_%i_hbox", layers[i]);
      hbox = gtk_object_get_data(GTK_OBJECT(dialog), key);
      pict_optionmenu = gtk_object_get_data(GTK_OBJECT(hbox), 
					    "pict_optionmenu");
            
      if( (w <= 96) && (h <= 63) )
	{
	  calc_type[i] = CALC_TI82;
	  calc_type[num_layers + i] = CALC_TI82;
	  pic_specs_dialog_update_pict_preview(dialog, layers[i], CALC_TI82);
	  gtk_option_menu_set_history(GTK_OPTION_MENU(pict_optionmenu), 0);
	}
      else if( (w <= 128) && (h <= 63) )
	{
	  calc_type[i] = CALC_TI85;
	  calc_type[num_layers + i] = CALC_TI85;
	  pic_specs_dialog_update_pict_preview(dialog, layers[i], CALC_TI85);
	  gtk_option_menu_set_history(GTK_OPTION_MENU(pict_optionmenu), 0);
	}
      else if( (w <= 239) && (h <= 103) )
	{
	  calc_type[i] = CALC_TI89;
	  calc_type[num_layers + i] = CALC_TI89;
	  pic_specs_dialog_update_pict_preview(dialog, layers[i], CALC_TI89);
	  gtk_option_menu_set_history(GTK_OPTION_MENU(pict_optionmenu), 0);
	}
      else
	{
	  // not supported !
	  return CALC_NONE;
	}
    }
  
  g_free(layers);
  
  gtk_widget_show(dialog);
  gtk_main();
  
  // Free tooltips
  gimp_help_free();

  // Check exit status, if user aborted, we'll cascade ...
  exit_okay = GPOINTER_TO_INT(gtk_object_get_data(GTK_OBJECT(dialog), 
						   "exit_status"));
  if(!exit_okay)
    calc_type = CALC_NONE;
  
  gtk_widget_destroy(dialog);
  
  return calc_type;
}

GimpPDBStatusType
SavePIC(gchar *filename, gint32 image)
{
  TiPicture pic;
  GimpPDBStatusType exit_state;
  gint *calc_type = NULL;

  D(("*** Saving TI Picture file %s\n", filename));

  if(pic_layers_too_big(image))
    {
      gimp_message(_("TI pictures cannot be higher or wider than 255 pixels.\n"));
      return GIMP_PDB_EXECUTION_ERROR;
    }

  if(gimp_image_base_type(image) != GIMP_RGB)
    {
      if(!gimp_image_convert_rgb(image))
	return GIMP_PDB_EXECUTION_ERROR;
    }

  // First, set up the picture specs dialog and show it
  if( (calc_type = pic_show_pict_dialog(image)) == NULL)
    return GIMP_PDB_CANCEL;

  if(interactive_pic)
    {
      guchar *temp_buf = g_strdup_printf(_("Saving %s:"), filename);
      gimp_progress_init(temp_buf);
      g_free(temp_buf);
    }

  // Okay, let's actually save the thing with the size the user specified.

  if((exit_state = pic_init(filename, &pic, calc_type) != GIMP_PDB_SUCCESS))
    return exit_state;

  D(("picture initialized ...\n"));

  pic_setup(&pic, image, calc_type);

  D(("picture data created ...\n"));

  if((exit_state = pic_save(&pic)) == GIMP_PDB_SUCCESS)
    {
      pic_sync_image_to_pic(image, calc_type);
    }

  D(("*** picture saved, exit status %i.\n\n", exit_state));

  pic_cleanup(&pic);

  return exit_state;
}
